@@ -19,6 +19,8 @@
 
 #include <cupt/common.hpp>
 #include <cupt/cache/relation.hpp>
+#include <cupt/packagename.hpp>
+#include <cupt/versionstring.hpp>
 
 #include <internal/common.hpp>
 #include <internal/parse.hpp>
@@ -26,12 +28,103 @@
 namespace cupt {
 namespace cache {
 
-bool Relation::__parse_versioned_info(const char* current, const char* end)
+namespace {
+
+const char* parseWhitespace(const char* start, const char* end)
+{
+	auto current = start;
+	while (current != end && *current == ' ')
+	{
+		++current;
+	}
+	return current;
+}
+
+template <typename CallbackT>
+const char* parseEnclosedWordList(const char* start, const char* end,
+		char openingChar, char closingChar, const char* failMessage, const CallbackT& callback)
+{
+	auto current = start;
+	if (current == end || *current != openingChar)
+	{
+		return current; // no architectures
+	}
+	++current;
+
+	const char* wordStart = nullptr;
+	bool foundClosingBracket = false;
+	while (current != end && !foundClosingBracket)
+	{
+		if (*current == ' ' || *current == closingChar)
+		{
+			if (wordStart)
+			{
+				callback(wordStart, current);
+				wordStart = nullptr;
+			}
+			foundClosingBracket = (*current == closingChar);
+		}
+		else
+		{
+			if (!wordStart)
+			{
+				wordStart = current;
+			}
+		}
+		++current;
+	}
+
+	if (!foundClosingBracket)
+	{
+		fatal2(failMessage, string(start, end));
+	}
+
+	return current;
+}
+
+}
+
+const char* Relation::p_parseVersionPart(const char* current, const char* end)
 {
 	// parse relation
-	if (current == end || current+1 == end /* version should at least have one character */)
+	if (current == end || *current != '(')
 	{
-		return false;
+		return current; // no version part detected
+	}
+	++current;
+
+	current = p_parseRelationSymbols(current, end);
+	if (!current)
+	{
+		return nullptr; // wrong symbols
+	}
+
+	current = parseWhitespace(current, end);
+
+	auto versionStringEnd = current;
+	while (versionStringEnd != end && *versionStringEnd != ')' && *versionStringEnd != ' ')
+	{
+		++versionStringEnd;
+	}
+	versionString.assign(current, versionStringEnd);
+	checkVersionString(versionString);
+
+	current = parseWhitespace(versionStringEnd, end);
+
+	if (current == end || *current != ')')
+	{
+		return nullptr; // no closing bracket detected
+	}
+	++current;
+
+	return current;
+}
+
+const char* Relation::p_parseRelationSymbols(const char* current, const char* end)
+{
+	if (current + 1 > end)
+	{
+		return nullptr;
 	}
 	switch (*current)
 	{
@@ -80,85 +173,60 @@ bool Relation::__parse_versioned_info(const char* current, const char* end)
 		}
 		break;
 		default:
-			return false;
+			return nullptr;
 	}
-	while (current != end && *current == ' ')
-	{
-		++current;
-	}
-	const char* versionStringEnd = current+1;
-	while (versionStringEnd != end && *versionStringEnd != ')' && *versionStringEnd != ' ')
-	{
-		++versionStringEnd;
-	}
-	if (versionStringEnd == end)
-	{
-		return false; // at least ')' after version string should be
-	}
-	versionString.assign(current, versionStringEnd);
-	checkVersionString(versionString);
-
-	current = versionStringEnd;
-	while (current != end && *current == ' ')
-	{
-		++current;
-	}
-	if (current == end || *current != ')')
-	{
-		return false;
-	}
-	++current;
-	while (current != end && *current == ' ')
-	{
-		++current;
-	}
-	return (current == end);
+	return current;
 }
 
-void Relation::__init(const char* start, const char* end)
+const char* Relation::p_parsePackagePart(const char* start, const char* end)
 {
 	const char* current;
+
 	consumePackageName(start, end, current);
-	if (current != start)
+	if (current == start) return nullptr;
+
+	packageName.assign(start, current);
+	relationType = Types::None;
+
+	if (current != end && *current == ':')
 	{
-		// package name is here
-		packageName.assign(start, current);
-	}
-	else
-	{
-		// no package name, bad
-		string unparsed(start, end);
-		fatal2(__("failed to parse a package name in the relation '%s'"), unparsed);
+		start = current+1;
+		consumePackageName(start, end, current);
+		if (current == start) return nullptr;
+
+		architecture.assign(start, current);
 	}
 
-	while (current != end && *current == ' ')
-	{
-		++current;
-	}
-	if (current != end && *current == '(')
-	{
-		++current;
-		// okay, here we should have a versoined info
-		if (!__parse_versioned_info(current, end))
-		{
-			string unparsed(start, end);
-			fatal2(__("failed to parse a version part in the relation '%s'"), unparsed);
-		}
-	}
-	else if (current == end)
-	{
-		relationType = Types::None;
-	}
-	else
-	{
-		// bad character in the middle of package name
-		fatal2(__("failed to parse a package name in the relation '%s'"), string(start, end));
-	}
+	return current;
 }
 
-Relation::Relation(pair< const char*, const char* > input)
+const char* Relation::__init(const char* start, const char* end)
 {
-	__init(input.first, input.second);
+	const char* current = p_parsePackagePart(start, end);
+	if (!current)
+	{
+		fatal2(__("failed to parse a package name in the relation '%s'"), string(start, end));
+	}
+	current = parseWhitespace(current, end);
+	current = p_parseVersionPart(current, end);
+	if (!current)
+	{
+		fatal2(__("failed to parse a version part in the relation '%s'"), string(start, end));
+	}
+	return parseWhitespace(current, end);
+}
+
+Relation::Relation(pair<const char*, const char*> input)
+{
+	if (__init(input.first, input.second) != input.second)
+	{
+		fatal2(__("failed to parse a suffix in the relation '%s'"), string(input.first, input.second));
+	};
+}
+
+Relation::Relation(pair<const char*, const char*> input, char const* * end)
+{
+	*end = __init(input.first, input.second);
 }
 
 Relation::~Relation()
@@ -167,6 +235,11 @@ Relation::~Relation()
 string Relation::toString() const
 {
 	string result = packageName;
+	if (!architecture.empty())
+	{
+		result += ':';
+		result += architecture;
+	}
 	if (relationType != Types::None)
 	{
 		// there is versioned info
@@ -217,27 +290,43 @@ bool Relation::operator==(const Relation& other) const
 
 const string Relation::Types::strings[] = { "<<", "=", ">>", "<=", ">=", "===" };
 
-void ArchitecturedRelation::__init(const char* start, const char* end)
+void ArchitecturedRelation::__init(const char* start, const char* suffixStart, const char* end)
 {
-	if (start == end)
+	auto current = p_parseArchitectures(suffixStart, end);
+	const char* prevCurrent;
+	do
 	{
-		return; // no architecture filters
-	}
-	if (*start != '[' || *(end-1) != ']')
-	{
-		fatal2(__("unable to parse architecture filters '%s'"), string(start, end));
-	}
-	++start;
-	--end;
+		prevCurrent = current;
+		current = parseWhitespace(current, end);
+		current = p_parseProfiles(current, end);
+	} while (current != prevCurrent);
 
-	architectureFilters = internal::split(' ', string(start, end));
+	if (current != end)
+	{
+		fatal2(__("failed to parse a suffix in the relation '%s'"), string(start, end));
+	};
 }
 
-ArchitecturedRelation::ArchitecturedRelation(
-		pair< const char*, const char* > input)
-	: Relation(std::make_pair(input.first, std::find(input.first, input.second, '[')))
+const char* ArchitecturedRelation::p_parseArchitectures(const char* start, const char* end)
 {
-	__init(std::find(input.first, input.second, '['), input.second);
+	return parseEnclosedWordList(start, end, '[', ']',
+			__("unable to parse architecture filters '%s'"),
+			[this](const char* a, const char* b){ architectureFilters.emplace_back(a, b); });
+}
+
+const char* ArchitecturedRelation::p_parseProfiles(const char* start, const char* end)
+{
+	return parseEnclosedWordList(start, end, '<', '>',
+			__("unable to parse build profiles '%s'"),
+			[](const char*, const char*){}); // TODO: save build-profiles
+}
+
+thread_local static const char* parentEnd;
+
+ArchitecturedRelation::ArchitecturedRelation(pair<const char*, const char*> input)
+	: Relation(input, &parentEnd)
+{
+	__init(input.first, parentEnd, input.second);
 }
 
 string ArchitecturedRelation::toString() const
