@@ -27,6 +27,7 @@
 #include <cupt/system/state.hpp>
 #include <cupt/file.hpp>
 #include <cupt/versionstring.hpp>
+#include <cupt/packagename.hpp>
 
 #include <internal/cacheimpl.hpp>
 #include <internal/filesystem.hpp>
@@ -55,7 +56,10 @@ void CacheImpl::processProvides(const string* packageNamePtr,
 {
 	auto callback = [this, &packageNamePtr](const char* tokenBeginIt, const char* tokenEndIt)
 	{
-		auto& sublist = this->canProvide[string(tokenBeginIt, tokenEndIt)];
+		const char* packageNameEndIt;
+		consumePackageName(tokenBeginIt, tokenEndIt, packageNameEndIt);
+
+		auto& sublist = this->canProvide[string(tokenBeginIt, packageNameEndIt)];
 		if (std::find(sublist.begin(), sublist.end(), packageNamePtr) == sublist.end())
 		{
 			sublist.push_back(packageNamePtr);
@@ -67,12 +71,12 @@ void CacheImpl::processProvides(const string* packageNamePtr,
 
 Package* CacheImpl::newBinaryPackage() const
 {
-	return new BinaryPackage(binaryArchitecture.get());
+	return new BinaryPackage();
 }
 
 Package* CacheImpl::newSourcePackage() const
 {
-	return new SourcePackage(binaryArchitecture.get());
+	return new SourcePackage();
 }
 
 Package* CacheImpl::preparePackage(unordered_map< string, vector< PrePackageRecord > >& pre,
@@ -91,14 +95,16 @@ Package* CacheImpl::preparePackage(unordered_map< string, vector< PrePackageReco
 		auto& package = target[packageName];
 		package.reset( (this->*packageBuilderMethod)() );
 
+		internal::VersionParseParameters versionInitParams;
+		versionInitParams.packageNamePtr = &packageName;
+		versionInitParams.binaryArchitecturePtr = binaryArchitecture.get();
+
 		vector< PrePackageRecord >& preRecord = preIt->second;
 		FORIT(preRecordIt, preRecord)
 		{
-			internal::VersionParseParameters versionInitParams;
 			versionInitParams.releaseInfo = preRecordIt->releaseInfoAndFile->first.get();
 			versionInitParams.file = preRecordIt->releaseInfoAndFile->second.get();
 			versionInitParams.offset = preRecordIt->offset;
-			versionInitParams.packageNamePtr = &packageName;
 			package->addEntry(versionInitParams);
 		}
 		preRecord.clear();
@@ -141,54 +147,66 @@ static inline void addSatisfyingPackageVersions(
 
 void CacheImpl::addRealPackageSatisfyingVersions(vector<const BinaryVersion*>* result, const Relation& relation) const
 {
-	const string& packageName = relation.packageName;
+	auto package = getBinaryPackage(relation.packageName);
 
-	size_t colonPosition = packageName.find(':');
-	if (colonPosition == string::npos)
+	if (relation.architecture.empty())
 	{
-		addSatisfyingPackageVersions<false>(result, relation, getBinaryPackage(packageName), *systemState);
+		addSatisfyingPackageVersions<false>(result, relation, package, *systemState);
 	}
-	else if (packageName.compare(colonPosition+1, string::npos, "any", 3) == 0)
+	else if (relation.architecture.compare(0, string::npos, "any", 3) == 0)
 	{
-		auto realPackageName = packageName.substr(0, colonPosition);
-		addSatisfyingPackageVersions<true>(result, relation, getBinaryPackage(realPackageName), *systemState);
+		addSatisfyingPackageVersions<true>(result, relation, package, *systemState);
 	}
 	// otherwise unsupported
 }
 
-void CacheImpl::addVirtualPackageSatisfyingVersions(vector<const BinaryVersion*>* result, const Relation& relation) const
-{
-	const string& packageName = relation.packageName;
+namespace {
 
-	// virtual package can only be considered if no relation sign is specified
+bool providesChunkMatchesRelation(const Relation& providesChunk, const Relation& relation)
+{
+	if (providesChunk.packageName != relation.packageName)
+	{
+		return false;
+	}
 	if (relation.relationType == Relation::Types::None)
 	{
-		// looking for reverse-provides
-		auto reverseProvidesIt = canProvide.find(packageName);
-		if (reverseProvidesIt != canProvide.end())
+		return true;
+	}
+	if (providesChunk.relationType != Relation::Types::Equal)
+	{
+		return false;
+	}
+	return relation.isSatisfiedBy(providesChunk.versionString);
+}
+
+}
+
+void CacheImpl::addVirtualPackageSatisfyingVersions(vector<const BinaryVersion*>* result, const Relation& relation) const
+{
+	auto reverseProvidesIt = canProvide.find(relation.packageName);
+	if (reverseProvidesIt != canProvide.end())
+	{
+		for (const auto& it: reverseProvidesIt->second)
 		{
-			for (const auto& it: reverseProvidesIt->second)
+			auto reverseProvidePackage = getBinaryPackage(*it);
+			if (!reverseProvidePackage)
 			{
-				auto reverseProvidePackage = getBinaryPackage(*it);
-				if (!reverseProvidePackage)
+				continue;
+			}
+			for (auto version: *reverseProvidePackage)
+			{
+				if (version->isInstalled() &&
+						systemState->getInstalledInfo(version->packageName)->isBroken())
 				{
 					continue;
 				}
-				for (auto version: *reverseProvidePackage)
+				for (const auto& providesChunk: version->provides)
 				{
-					if (version->isInstalled() &&
-							systemState->getInstalledInfo(version->packageName)->isBroken())
+					if (providesChunkMatchesRelation(providesChunk, relation))
 					{
-						continue;
-					}
-					for (const auto& realProvidesPackageName: version->provides)
-					{
-						if (realProvidesPackageName == packageName)
-						{
-							// ok, this particular version does provide this virtual package
-							result->push_back(version);
-							break;
-						}
+						// ok, this particular version does provide this virtual package
+						result->push_back(version);
+						break;
 					}
 				}
 			}
@@ -727,7 +745,8 @@ ssize_t CacheImpl::computePin(const Version* version, const BinaryPackage* binar
 		for (const auto& otherVersion: *binaryPackage)
 		{
 			if (otherVersion == version) continue;
-			if (versionstring::sameOriginal(otherVersion->versionString, installedVersionString))
+			if (getOriginalVersionString(otherVersion->versionString).equal(
+					getOriginalVersionString(installedVersionString)))
 			{
 				auto otherPin = getPin(otherVersion, [&binaryPackage]() { return binaryPackage; });
 				if (otherPin > result) result = otherPin;
@@ -775,12 +794,12 @@ string CacheImpl::getLocalizedDescription(const BinaryVersion* version) const
 
 void CacheImpl::parseExtendedStates()
 {
-	// we are parsing duals like:
+	// we are mainly parsing duals like:
 
 	// Package: perl
 	// Auto-Installed: 1
 
-	// but, rarely another fields may be present, we need to ignore them
+	// but another fields may be present
 
 	try
 	{
@@ -794,42 +813,7 @@ void CacheImpl::parseExtendedStates()
 			return;
 		}
 
-		internal::TagParser parser(&file);
-		internal::TagParser::StringRange tagName, tagValue;
-
-		while (parser.parseNextLine(tagName, tagValue) && !file.eof())
-		{
-			if (!tagName.equal("Package", 7))
-			{
-				fatal2(__("wrong tag: expected 'Package', got '%s'"), tagName.toString());
-			}
-
-			string packageName = tagValue.toString();
-
-			bool valueFound = false;
-			while (parser.parseNextLine(tagName, tagValue))
-			{
-				if (tagName.equal(BUFFER_AND_SIZE("Auto-Installed")))
-				{
-					valueFound = true;
-					if (tagValue.equal(BUFFER_AND_SIZE("1")))
-					{
-						// adding to storage
-						extendedInfo.automaticallyInstalled.insert(packageName);
-					}
-					else if (!tagValue.equal(BUFFER_AND_SIZE("0")))
-					{
-						fatal2(__("bad value '%s' (should be 0 or 1) for the package '%s'"),
-								tagValue.toString(), packageName);
-					}
-				}
-			}
-
-			if (!valueFound)
-			{
-				fatal2(__("didn't found the 'Auto-Installed' tag for the package '%s'"), packageName);
-			}
-		}
+		p_parseExtendedStatesContent(file);
 	}
 	catch (Exception&)
 	{
@@ -837,6 +821,42 @@ void CacheImpl::parseExtendedStates()
 	}
 }
 
+void CacheImpl::p_parseExtendedStatesContent(File& contentFile)
+{
+	internal::TagParser parser(&contentFile);
+	internal::TagParser::StringRange tagName, tagValue;
+
+	while (parser.parseNextLine(tagName, tagValue) && !contentFile.eof())
+	{
+		if (!tagName.equal("Package", 7))
+		{
+			fatal2(__("wrong tag: expected 'Package', got '%s'"), tagName.toString());
+		}
+
+		string packageName = tagValue.toString();
+		auto& rawRecord = extendedInfo.raw[packageName];
+
+		while (parser.parseNextLine(tagName, tagValue))
+		{
+			rawRecord[tagName.toString()] = tagValue.toString();
+
+			if (tagName.equal(BUFFER_AND_SIZE("Auto-Installed")))
+			{
+				if (tagValue.equal(BUFFER_AND_SIZE("1")))
+				{
+					// adding to storage
+					extendedInfo.automaticallyInstalled.insert(packageName);
+				}
+				else if (!tagValue.equal(BUFFER_AND_SIZE("0")))
+				{
+					fatal2(__("bad value '%s' (should be 0 or 1) for the package '%s'"),
+							tagValue.toString(), packageName);
+				}
+			}
+		}
+	}
+}
+
 vector< const BinaryVersion* >
 CacheImpl::getSatisfyingVersionsNonCached(const RelationExpression& relationExpression) const
 {
