@@ -20,6 +20,8 @@
 #include <fcntl.h>
 #include <time.h>
 #include <unistd.h>
+#include <clocale>
+#include <cstdlib>
 
 #include <algorithm>
 
@@ -137,7 +139,7 @@ void SnapshotsWorker::__do_repacks(const vector< string >& installedPackageNames
 				}
 				const string& badFilename = files[0];
 				auto goodFilename = format2("%s_%s_%s.deb", packageName,
-						versionstring::getOriginal(version->versionString), architecture);
+						getOriginalVersionString(version->versionString).toStdString(), architecture);
 
 				if (!fs::move(badFilename, goodFilename))
 				{
@@ -215,22 +217,42 @@ void SnapshotsWorker::__create_release_file(const string& temporarySnapshotDirec
 	createTextFile(path, lines, _logger, simulating);
 }
 
+static const char* findProhibitedCharactersInSnapshotName(const string& name)
+{
+	auto isProhibited = [](char c)
+	{
+		if (!isascii(c)) return true;
+		if (isdigit(c)) return false;
+		if (islower(c)) return false;
+		if (c == '-') return false;
+		if (c == '_') return false;
+		return true;
+	};
+
+	for (const auto& c: name)
+	{
+		if (isProhibited(c))
+		{
+			return &c;
+		}
+	}
+	return nullptr;
+}
+
 void checkSnapshotName(const Snapshots& snapshots, const string& name)
 {
 	if (name.empty())
 	{
-		fatal2(__("the system snapshot name cannot be empty"));
+		fatal2(__("the system snapshot name '' cannot be empty"));
 	}
 	if (name[0] == '.')
 	{
 		fatal2(__("the system snapshot name '%s' cannot start with a '.'"), name);
 	}
-	for (auto c: name)
+
+	if (auto badChar = findProhibitedCharactersInSnapshotName(name))
 	{
-		if (isspace(c))
-		{
-			fatal2(__("the system snapshot name '%s' cannot contain a whitespace character '%c'"), name, c);
-		}
+		fatal2(__("the system snapshot name '%s' cannot contain a character '%c'"), name, *badChar);
 	}
 
 	{
